import pytest

from threedi_schema.domain.custom_types import clean_csv_string, clean_csv_table


@pytest.mark.parametrize(
    "value",
    [
        "1,2,3",
        "1, 2, 3 ",
        "1,\t2,3",
        "1,\r2,3 ",
        "1,\n2,3 ",
        "1,  2,3",
        "1,  2  ,3",
        " 1,2,3 ",
        "\n1,2,3",
        "\t1,2,3",
        "\r1,2,3",
        "1,2,3\t",
        "1,2,3\n",
        "1,2,3\r",
    ],
)
def test_clean_csv_string(value):
    assert clean_csv_string(value) == "1,2,3"


def test_clean_csv_string_with_whitespace():
    assert clean_csv_string("1,2 3,4") == "1,2 3,4"


@pytest.mark.parametrize(
    "value",
    [
        "1,2,3\n4,5,6",
        "1,2,3\r\n4,5,6",
        "\n1,2,3\n4,5,6",
        "1,2,3\n4,5,6\n",
    ],
)
def test_clean_csv_table(value):
    assert clean_csv_table(value) == "1,2,3\n4,5,6"


@pytest.mark.parametrize(
    "value", [" ", "0 1", "3;5", "foo", "1,2\n3,", ",2", ",2\n3,4"]
)
def test_clean_csv_table_no_fail(value):
    clean_csv_table(value)



"""Upgrade settings in schema

Revision ID: 0222
Revises:
Create Date: 2024-03-04 10:06

"""
import uuid

from typing import Dict, List, Tuple
from pathlib import Path

import sqlalchemy as sa
from alembic import op
from sqlalchemy import Boolean, Column, Float, Integer, String
from sqlalchemy.orm import declarative_base

from threedi_schema.migrations.utils import drop_conflicting

# revision identifiers, used by Alembic.
revision = "0222"
down_revision = "0221"
branch_labels = None
depends_on = None

Base = declarative_base()

data_dir = Path(__file__).parent / "data"


# (source table, destination table)
RENAME_TABLES = [
    ("v2_aggregation_settings", "aggregation_settings"),
    ("v2_groundwater", "groundwater"),
    ("v2_interflow", "interflow"),
    ("v2_numerical_settings", "numerical_settings"),
    ("v2_simple_infiltration", "simple_infiltration"),
    ("v2_vegetation_drag", "vegetation_drag_2d"),
    ("v2_global_settings", "model_settings")
]

# (old name, new name)
RENAME_COLUMNS = {
    "aggregation_settings":
        [
            ("timestep", "interval"),
        ],
    "groundwater":
        [
            ("groundwater_hydro_connectivity", "groundwater_hydraulic_conductivity"),
            ("groundwater_hydro_connectivity_file", "groundwater_hydraulic_conductivity_file"),
            ("groundwater_hydro_connectivity_type", "groundwater_hydraulic_conductivity_aggregation"),
            ("groundwater_impervious_layer_level_type", "groundwater_impervious_layer_level_aggregation"),
            ("infiltration_decay_period_type", "infiltration_decay_period_aggregation"),
            ("initial_infiltration_rate_type", "initial_infiltration_rate_aggregation"),
            ("phreatic_storage_capacity_type", "phreatic_storage_capacity_aggregation"),
            ("equilibrium_infiltration_rate_type", "equilibrium_infiltration_rate_aggregation"),
        ],
    "numerical_settings":
        [
            ("frict_shallow_water_correction", "friction_shallow_water_depth_correction"),
            ("thin_water_layer_definition", "limiter_slope_thin_water_layer"),
            ("limiter_grad_1d", "limiter_waterlevel_gradient_1d"),
            ("limiter_grad_2d", "limiter_waterlevel_gradient_2d"),
            ("max_degree", "max_degree_gauss_seidel"),
            ("max_nonlin_iterations", "max_non_linear_newton_iterations"),
            ("minimum_friction_velocity", "min_friction_velocity"),
            ("minimum_surface_area", "min_surface_area"),
            ("integration_method", "time_integration_method"),
            ("use_of_nested_newton", "use_nested_newton"),
            ("precon_cg", "use_preconditioner_cg"),
        ],
    "simple_infiltration":
        [
            ("max_infiltration_capacity", "max_infiltration_volume"),
            ("max_infiltration_capacity_file", "max_infiltration_volume_file"),
            ("infiltration_rate", "infiltration_rate")
        ],
    "model_settings":
        [
            ("dist_calc_points", "calculation_point_distance_1d"),
            ("frict_avg", "friction_averaging"),
            ("frict_coef", "friction_coefficient"),
            ("frict_coef_file", "friction_coefficient_file"),
            ("frict_type", "friction_type"),
            ("manhole_storage_area", "manhole_aboveground_storage_area"),
            ("grid_space", "minimum_cell_size"),
            ("kmax", "nr_grid_levels"),
            ("table_step_size", "minimum_table_step_size"),
        ],
}

ADD_COLUMNS = [
    ("numerical_settings", Column("flooding_threshold", Float)),
    ("model_settings", Column("use_groundwater_flow", Boolean)),
    ("model_settings", Column("use_groundwater_storage", Boolean)),
    ("model_settings", Column("use_structure_control", Boolean)),
    ("model_settings", Column("use_simple_infiltration", Boolean)),
    ("model_settings", Column("use_vegetation_drag_2d", Boolean)),
    ("model_settings", Column("use_interflow", Boolean)),
    ("model_settings", Column("use_interception", Boolean)),
]


ADD_TABLES = {
    "initial_conditions":
        [Column("initial_groundwater_level", Float),
         Column("initial_groundwater_level_aggregation", Integer),
         Column("initial_groundwater_level_file", String),
         Column("initial_water_level", Float),
         Column("initial_water_level_aggregation", Integer),
         Column("initial_water_level_file", String)],
    "interception":
        [Column("interception", Float),
         Column("interception_file", String)],
    "physical_settings":
        [Column("use_advection_1d", Integer),
         Column("use_advection_2d", Integer)],
    "simulation_template_settings":
        [Column("name", String), Column("use_0d_inflow", Integer)],
    "time_step_settings":
        [Column("max_time_step", Float),
         Column("min_time_step", Float),
         Column("output_time_step", Float),
         Column("time_step", Float),
         Column("use_time_step_stretch", Boolean)],
}

REMOVE_COLUMNS = [
    ("aggregation_settings", ["global_settings_id", "var_name"]),
    ("groundwater", ["display_name"]),
    ("interflow", ["display_name"]),
    ("model_settings", ["nr_timesteps",
                        "start_time",
                        "start_date",
                        "guess_dams",
                        "dem_obstacle_detection",
                        "dem_obstacle_height",
                        "wind_shielding_file"]),
    ("simple_infiltration", ["display_name"]),
    ("vegetation_drag_2d", ["display_name"]),
]

COPY_FROM_GLOBAL = {
    "simulation_template_settings": [
        ("name", "name"),
        ("use_0d_inflow", "use_0d_inflow")
    ],
    "time_step_settings": [
        ("maximum_sim_time_step", "max_time_step"),
        ("minimum_sim_time_step", "min_time_step"),
        ("output_time_step", "output_time_step"),
        ("sim_time_step", "time_step"),
        ("timestep_plus", "use_time_step_stretch"),
    ],
    "initial_conditions": [
        ("initial_groundwater_level", "initial_groundwater_level"),
        ("initial_groundwater_level_type", "initial_groundwater_level_aggregation"),
        ("initial_groundwater_level_file", "initial_groundwater_level_file"),
        ("initial_waterlevel", "initial_water_level"),
        ("water_level_ini_type", "initial_water_level_aggregation"),
        ("initial_waterlevel_file", "initial_water_level_file"),
    ],
    "physical_settings": [
        ("advection_1d", "use_advection_1d"),
        ("advection_2d", "use_advection_2d"),
    ],
    "interception": [
        ("interception_global", "interception"),
        ("interception_file", "interception_file"),
    ],
}

# Columns set to True if a mapping between use_* and settings table exists
# (boolean column, setting id, setting table)
GLOBAL_SETTINGS_ID_TO_BOOL = [
    ("use_groundwater_storage", "groundwater_settings_id", "groundwater"),
    ("use_interflow", "interflow_settings_id", "interflow"),
    ("use_structure_control", "control_group_id", "v2_control_group"),
    ("use_simple_infiltration", "simple_infiltration_settings_id", "simple_infiltration"),
    ("use_vegetation_drag_2d", "vegetation_drag_settings_id", "vegetation_drag_2d"),
]


def rename_tables(table_sets: List[Tuple[str, str]]):
    # no checks for existence are done, this will fail if a source table doesn't exist
    for src_name, dst_name in table_sets:
        op.rename_table(src_name, dst_name)


def rename_columns(table_name: str, columns: List[Tuple[str, str]]):
    # no checks for existence are done, this will fail if table or any source column doesn't exist
    with op.batch_alter_table(table_name) as batch_op:
        for src_name, dst_name in columns:
            batch_op.alter_column(src_name, new_column_name=dst_name)


def make_all_columns_nullable(table_name, id_name: str = 'id'):
    # no checks for existence are done, this will fail if table doesn't exist
    connection = op.get_bind()
    table = sa.Table(table_name, sa.MetaData(), autoload_with=connection)
    with op.batch_alter_table(table_name) as batch_op:
        for column in table.columns:
            if column.name == id_name:
                continue
            batch_op.alter_column(column_name=column.name, nullable=True)


def create_new_tables(new_tables: Dict[str, sa.Column]):
    # no checks for existence are done, this will fail if any table already exists
    for table_name, columns in new_tables.items():
        op.create_table(table_name, sa.Column("id", sa.Integer(), primary_key=True),
                        *columns)


def add_columns_to_tables(table_columns: List[Tuple[str, Column]]):
    # no checks for existence are done, this will fail if any column already exists
    for dst_table, col in table_columns:
        with op.batch_alter_table(dst_table) as batch_op:
            batch_op.add_column(col)


def move_multiple_values_to_empty_table(src_table: str, dst_table: str, columns: List[str]):
    # move values from one table to another
    # no checks for existence are done, this will fail if any table or column doesn't exist
    dst_cols = ', '.join(f'"{dst}"' for _, dst in columns)
    src_cols = ', '.join(src for src, _ in columns)
    op.execute(sa.text(f'INSERT INTO {dst_table} ({dst_cols}) SELECT {src_cols} FROM {src_table}'))
    remove_columns_from_table(src_table, [src for src, _ in columns])


def move_values_to_table(src_table: str, src_col: str, dst_table: str, dst_col: str):
    op.execute(f'UPDATE {dst_table} SET {dst_col} = (SELECT {src_col} FROM {src_table} ORDER BY id LIMIT 1)')
    remove_columns_from_table(src_table, [src_col])


def remove_columns_from_table(table_name: str, columns: List[str]):
    # no checks for existence are done, this will fail if any table or column doesn't exist
    with op.batch_alter_table(table_name) as batch_op:
        for column in columns:
            batch_op.drop_column(column)


def set_use_from_settings_id():
    for settings_col, settings_id, settings_table in GLOBAL_SETTINGS_ID_TO_BOOL:
        if settings_col != '':
            # set boolean 'use_*' in model_settings if a relationship exists
            op.execute(f"UPDATE model_settings SET {settings_col} = TRUE WHERE {settings_id} IS NOT NULL;")
        # remove all settings rows, exact for the one matching
        # delete_all_but_matching_id(settings_table, settings_id)
        op.execute(
            f"DELETE FROM {settings_table} WHERE NOT EXISTS "
            f"(SELECT 1 FROM model_settings WHERE model_settings.{settings_id} IS NOT NULL);")
        # command above doesn't catch id = None, so an extra command is needed for those cases
        op.execute(
            f"DELETE FROM {settings_table} WHERE id != "
            f"(SELECT {settings_id} FROM model_settings);")

    # set use_groundwater_flow
    sql = """
        UPDATE model_settings
        SET use_groundwater_flow = CASE
            WHEN (SELECT groundwater_hydraulic_conductivity FROM groundwater LIMIT 1) IS NOT NULL OR (SELECT groundwater_hydraulic_conductivity_file FROM groundwater LIMIT 1) IS NOT NULL THEN 1
            ELSE 0
        END;
        """
    op.execute(sql)


def drop_columns(table, columns):
    with op.batch_alter_table(table) as batch_op:
        for column in columns:
            batch_op.drop_column(column)


def set_use_inteception():
    # Set use_interception based on interception and interception_file values
    op.execute(sa.text("""
    UPDATE model_settings
    SET use_interception = (
        SELECT 
            CASE WHEN
                interception.interception IS NOT NULL  
                OR 
                (interception.interception_file IS NOT NULL 
                AND interception.interception_file != '') 
                THEN 1
                ELSE 0
            END
        FROM interception
    );    
    """))

    op.execute(sa.text("""
    DELETE FROM interception 
    WHERE (interception IS NULL OR interception = '')
    AND (interception_file IS NULL OR interception_file = '');    
    """))


def delete_all_but_matching_id(table, settings_id):
    op.execute(f"DELETE FROM {table} WHERE id NOT IN (SELECT {settings_id} FROM model_settings);")


def delete_all_but_first_row(table):
    op.execute(f"DELETE FROM {table} WHERE id NOT IN ("
               f"SELECT id FROM "
               f"(SELECT id FROM {table} ORDER BY id LIMIT 1) AS subquery);")


def correct_raster_paths():
    # Replace paths to raster files with only the file name
    raster_paths = [
        ("model_settings", "dem_file"),
        ("model_settings", "friction_coefficient_file"),
        ("interception", "interception_file"),
        ("interflow", "porosity_file"),
        ("interflow", "hydraulic_conductivity_file"),
        ("simple_infiltration", "infiltration_rate_file"),
        ("simple_infiltration", "max_infiltration_volume_file"),
        ("groundwater", "groundwater_impervious_layer_level_file"),
        ("groundwater", "phreatic_storage_capacity_file"),
        ("groundwater", "initial_infiltration_rate_file"),
        ("groundwater", "equilibrium_infiltration_rate_file"),
        ("groundwater", "infiltration_decay_period_file"),
        ("groundwater", "groundwater_hydraulic_conductivity_file"),
        ("initial_conditions", "initial_water_level_file"),
        ("initial_conditions", "initial_groundwater_level_file"),
        ("vegetation_drag_2d", "vegetation_height_file"),
        ("vegetation_drag_2d", "vegetation_stem_count_file"),
        ("vegetation_drag_2d", "vegetation_stem_diameter_file"),
        ("vegetation_drag_2d", "vegetation_drag_coefficient_file"),
    ]
    conn = op.get_bind()
    for table, col in raster_paths:
        result = conn.execute(sa.text(f"SELECT id, {col} FROM {table} WHERE {col} IS NOT NULL;")).fetchall()
        # model_settings only has one row, so we can just grab that one
        if len(result) == 1:
            id, file_path = result[0]
            if isinstance(file_path, str) and len(file_path) > 0:
                # replace backslash in windows paths because pathlib doesn't handle relative windows paths
                file_path = file_path.replace('\\', '/')
                file = Path(file_path).name
                op.execute(
                    sa.text(f"UPDATE {table} SET {col} = :new_value WHERE id = :row_id")
                    .bindparams(new_value=file, row_id=id)
                )


def remove_columns_from_copied_tables(table_name: str, rem_columns: List[str]):
    # sqlite 3.27 doesn't support `ALTER TABLE ... DROP COLUMN`
    # So we create a temp table, copy the columns we want to keep and remove the old table
    # Retrieve columns
    connection = op.get_bind()
    all_columns = connection.execute(sa.text(f"PRAGMA table_info('{table_name}')")).fetchall()
    col_names = [col[1] for col in all_columns if col[1] not in rem_columns]
    col_types = [col[2] for col in all_columns if col[1] not in rem_columns]
    cols = (['id INTEGER PRIMARY KEY NOT NULL'] +
            [f'{cname} {ctype}' for cname, ctype in zip(col_names, col_types) if cname != 'id'])
    # Create new table, insert data, drop original and rename to table_name
    temp_name = f'_temp_222_{uuid.uuid4().hex}'
    op.execute(sa.text(f"CREATE TABLE {temp_name} ({','.join(cols)});"))
    op.execute(sa.text(f"INSERT INTO {temp_name} ({','.join(col_names)}) SELECT {','.join(col_names)} FROM {table_name}"))
    op.execute(sa.text(f"DROP TABLE {table_name};"))
    op.execute(sa.text(f"ALTER TABLE {temp_name} RENAME TO {table_name};"))


def set_flow_variable_values():
    flow_var_dict = {'wet_cross-section': 'wet_cross_section',
                     'waterlevel': 'water_level'}
    cases = '\n'.join([f"WHEN '{key}' THEN '{val}'" for key, val in flow_var_dict.items()])
    query = f"""
    UPDATE aggregation_settings SET flow_variable = CASE flow_variable
    {cases} 
    ELSE flow_variable
    END"""
    op.execute(sa.text(query))


def upgrade():
    op.get_bind()
    # Only use first row of global settings
    delete_all_but_first_row("v2_global_settings")
    # Remove existing tables (outside of the specs) that conflict with new table names
    drop_conflicting(op, list(ADD_TABLES.keys()) + [new_name for _, new_name in RENAME_TABLES])
    rename_tables(RENAME_TABLES)
    # rename columns in renamed tables
    for table_name, columns in RENAME_COLUMNS.items():
        rename_columns(table_name, columns)
    # make all columns in renamed tables, except id, nullable
    for _, table_name in RENAME_TABLES:
        make_all_columns_nullable(table_name)
    # create new tables
    create_new_tables(ADD_TABLES)
    # add empty columns to tables
    add_columns_to_tables(ADD_COLUMNS)
    # copy data from model_settings to new tables and columns
    for dst_table, columns in COPY_FROM_GLOBAL.items():
        move_multiple_values_to_empty_table("model_settings", dst_table, columns)
    # copy data from model_settings to existing table
    move_values_to_table('model_settings', 'flooding_threshold', 'numerical_settings', 'flooding_threshold')
    # set several 'use' columns based on other settings
    set_use_from_settings_id()
    set_use_inteception()
    # keep first row of settings tables that have no explicit mapping
    delete_all_but_matching_id("numerical_settings", "numerical_settings_id")
    # drop unused id columns in model_settings
    unused_cols = [settings_id for _, settings_id, _ in GLOBAL_SETTINGS_ID_TO_BOOL] + ["numerical_settings_id"]
    drop_columns('model_settings', unused_cols)
    # remove relative path prefix from raster paths
    correct_raster_paths()
    # change flow_variable values to new naming scheme
    set_flow_variable_values()
    # remove columns from tables that are copied
    for table, columns in REMOVE_COLUMNS:
        remove_columns_from_copied_tables(table, columns)


def downgrade():
    # Not implemented on purpose
    raise NotImplementedError("Downgrade back from 0.3xx is not supported")



from sqlalchemy import func, inspect, text

__all__ = ["ensure_spatial_indexes"]


def create_spatial_index(connection, column):
    """
    Create spatial index for given column.
    Note that this will fail if the spatial index already exists!
    """
    idx_name = f"{column.table.name}_{column.name}"
    try:
        connection.execute(func.gpkgAddSpatialIndex(column.table.name, column.name))
    except Exception as e:
        raise RuntimeError(
            f"Spatial index creation for {idx_name} failed with error {e}"
        )
    return True


def get_missing_spatial_indexes(engine, models):
    """
    Collect all rtree tables that should exist
    There can only be one geometry column per table and we assume any geometry column is named geom
    """
    inspector = inspect(engine)
    table_names = inspector.get_table_names()
    return [
        model
        for model in models
        if "geom" in model.__table__.columns
        and f"rtree_{model.__table__.name}_geom" not in table_names
    ]


def ensure_spatial_indexes(engine, models):
    """Ensure presence of spatial indexes for all geometry columns"""
    created = False
    no_spatial_index_models = get_missing_spatial_indexes(engine, models)
    with engine.connect() as connection:
        with connection.begin():
            for model in no_spatial_index_models:
                created &= create_spatial_index(
                    connection, model.__table__.columns["geom"]
                )
            if created:
                connection.execute(text("VACUUM"))



"""Upgrade structure control

Revision ID: 0224
Revises:
Create Date: 2024-06-30 14:50

"""
import uuid
import warnings
from typing import Dict, List, Tuple

import sqlalchemy as sa
from alembic import op
from sqlalchemy import Boolean, Column, Float, Integer, String, Text
from sqlalchemy.orm import declarative_base

from threedi_schema.domain.custom_types import Geometry
from threedi_schema.migrations.utils import drop_conflicting, drop_geo_table

# revision identifiers, used by Alembic.
revision = "0224"
down_revision = "0223"
branch_labels = None
depends_on = None

Base = declarative_base()

DEL_TABLES = ['v2_control',
              'v2_control_delta',
              'v2_control_group',
              'v2_control_measure_group',
              'v2_control_measure_map',
              'v2_control_pid',
              'v2_control_timed']

RENAME_TABLES = [
    ("v2_control_memory", "memory_control"),
    ("v2_control_table", "table_control"),
]

ADD_COLUMNS = [
    ("memory_control", Column("tags", Text)),
    ("memory_control", Column("code", Text)),
    ("memory_control", Column("display_name", Text)),
    ("memory_control", Column("action_value_1", Float)),
    ("memory_control", Column("action_value_2", Float)),
    ("table_control", Column("tags", Text)),
    ("table_control", Column("code", Text)),
    ("table_control", Column("display_name", Text)),
    ('simulation_template_settings', Column('use_structure_control', Boolean)),
    ("control_measure_location", Column("geom", Geometry("POINT"), nullable=False)),
    ("control_measure_map", Column("geom", Geometry("LINESTRING"), nullable=False)),
    ("memory_control", Column("geom", Geometry("POINT"), nullable=False)),
    ("table_control", Column("geom", Geometry("POINT"), nullable=False))
]

ADD_TABLES = {
    "control_measure_location": [
        Column("connection_node_id", Integer),
        Column("measure_variable", Text, server_default="water_level"),
        Column("tags", Text),
        Column("code", Text),
        Column("display_name", Text),
    ],
    "control_measure_map": [
        Column("weight", Float),
        Column("control_measure_location_id", Integer),
        Column("control_id", Integer),
        Column("control_type", Text),
        Column("tags", Text),
        Column("code", Text),
        Column("display_name", Text),
    ],
}


class InvalidConnectionNode(UserWarning):
    pass


def create_new_tables(new_tables: Dict[str, sa.Column]):
    # no checks for existence are done, this will fail if any table already exists
    for table_name, columns in new_tables.items():
        op.create_table(table_name, sa.Column("id", sa.Integer(), primary_key=True),
                        *columns)


def rename_tables(table_sets: List[Tuple[str, str]]):
    # no checks for existence are done, this will fail if a source table doesn't exist
    for src_name, dst_name in table_sets:
        op.rename_table(src_name, dst_name)


def add_columns_to_tables(table_columns: List[Tuple[str, Column]]):
    # no checks for existence are done, this will fail if any column already exists
    for dst_table, col in table_columns:
        if isinstance(col.type, Geometry):
            add_geometry_column(dst_table, col)
        else:
            with op.batch_alter_table(dst_table) as batch_op:
                batch_op.add_column(col)


def make_all_columns_nullable(table_name, id_name: str = 'id'):
    # no checks for existence are done, this will fail if table doesn't exist
    connection = op.get_bind()
    table = sa.Table(table_name, sa.MetaData(), autoload_with=connection)
    with op.batch_alter_table(table_name) as batch_op:
        for column in table.columns:
            if column.name == id_name:
                continue
            batch_op.alter_column(column_name=column.name, nullable=True)


def move_setting(src_table: str, src_col: str, dst_table: str, dst_col: str):
    # This only works for tables with one row, such as settings tables
    op.execute(sa.text(f'UPDATE {dst_table} SET {dst_col} = (SELECT {src_col} FROM {src_table} ORDER BY id LIMIT 1)'))
    remove_column_from_table(src_table, src_col)


def remove_invalid_rows_from_v2_control_measure_map():
    conn = op.get_bind()
    where_clause = """
    WHERE object_id NOT IN (
        SELECT id FROM v2_connection_nodes WHERE the_geom IS NOT NULL
    )
    """
    invalid_rows = conn.execute(sa.text(f"SELECT id FROM v2_control_measure_map {where_clause};")).fetchall()
    if len(invalid_rows) > 0:
        op.execute(sa.text(f"DELETE FROM v2_control_measure_map {where_clause};"))
        msg = (f"Could not migrate the following rows from v2_control_measure_map because "
               f"they are linked to connection nodes that do not exist"
               f"or have no geometry: {invalid_rows}")
        warnings.warn(msg, InvalidConnectionNode)



def remove_orphan_control(table_name):
    query = f"DELETE FROM {table_name} WHERE id NOT IN (SELECT control_id FROM control_measure_map);"
    op.execute(sa.text(query))


def populate_control_measure_map():
    query = """
    INSERT INTO control_measure_map (control_measure_location_id, weight, control_id, control_type)
    SELECT 
        v2_control_measure_map.id, 
        v2_control_measure_map.weight, 
        v2_control.control_id, 
        v2_control.control_type
    FROM 
        v2_control_measure_map
    JOIN 
        v2_control
    ON 
        v2_control_measure_map.measure_group_id = v2_control.measure_group_id;
    """
    op.execute(sa.text(query))


def add_control_geometries(control_name):
    targets = ['v2_channel', 'v2_pipe', 'v2_orifice', 'v2_culvert', 'v2_weir', 'v2_pumpstation']
    for target_type in targets:
        if target_type == 'v2_culvert':
            geom_query = f"""
                SELECT ST_Centroid(object.the_geom)
                FROM v2_culvert AS object
                JOIN v2_cross_section_definition AS def ON object.cross_section_definition_id = def.id
            """
        elif target_type == 'v2_pumpstation':
            geom_query = f"""
                SELECT start_node.the_geom
                FROM {target_type} AS object   
                JOIN v2_connection_nodes AS start_node ON object.connection_node_start_id = start_node.id
            """
        else:
            geom_query = f"""
                SELECT ST_Centroid(MakeLine(start_node.the_geom, end_node.the_geom))
                FROM {target_type} AS object
            """
            if target_type in ['v2_orifice', 'v2_weir']:
                geom_query += """
                JOIN v2_cross_section_definition AS def ON object.cross_section_definition_id = def.id
                """
            geom_query += """
            JOIN v2_connection_nodes AS start_node ON object.connection_node_start_id = start_node.id
            JOIN v2_connection_nodes AS end_node ON object.connection_node_end_id = end_node.id         
            """
        query = f"""
            UPDATE
                {control_name}
            SET geom = ({geom_query}
            WHERE object.id = {control_name}.target_id
            )
            WHERE {control_name}.target_type = '{target_type}';                
        """
        op.execute(sa.text(query))


def get_global_srid():
    conn = op.get_bind()
    use_0d_inflow = conn.execute(sa.text("SELECT use_0d_inflow FROM simulation_template_settings LIMIT 1")).fetchone()
    if use_0d_inflow is not None:
        srid = conn.execute(sa.text("SELECT epsg_code FROM model_settings LIMIT 1")).fetchone()
        if (srid is not None) and (srid[0] is not None):
            return srid[0]
    return 28992


def set_geom_for_control_measure_map():
    srid = get_global_srid()
    for control in ['memory', 'table']:
        control_table = f'{control}_control'
        query = f"""
        UPDATE
            control_measure_map
        SET
            geom = (
                SELECT CASE
                    WHEN ST_Equals(cml.geom, tc.geom) THEN
                        -- Transform to EPSG:4326 for the projection, then back to the original SRID
                        MakeLine(
                            cml.geom,
                            ST_Transform(
                                ST_Translate(
                                    ST_Transform(tc.geom, {srid}),
                                    0, 1, 0
                                ),
                                4326
                            )                   
                        )
                    ELSE
                        MakeLine(cml.geom, tc.geom)
                    END                                           
                FROM 
                    {control_table} AS tc
                JOIN 
                    control_measure_map AS cmm ON cmm.control_id = tc.id
                JOIN 
                    control_measure_location AS cml ON cmm.control_measure_location_id = cml.id
                WHERE 
                    tc.id = control_measure_map.control_id
                )    
            WHERE
                EXISTS (
                    SELECT 1
                    FROM 
                        {control_table} AS tc
                    WHERE 
                        tc.id = control_measure_map.control_id
                );                                     
        """
        op.execute(sa.text(query))


def add_geometry_column(table: str, geocol: Column):
    # Adding geometry columns via alembic doesn't work
    # https://postgis.net/docs/AddGeometryColumn.html
    geotype = geocol.type
    query = (
        f"SELECT AddGeometryColumn('{table}', '{geocol.name}', {geotype.srid}, '{geotype.geometry_type}', 'XY', 0);")
    op.execute(sa.text(query))


def populate_control_measure_location():
    # copy data from
    query = """
    INSERT INTO control_measure_location (id, connection_node_id)
    SELECT 
        v2_control_measure_map.id, 
        v2_control_measure_map.object_id
    FROM 
        v2_control_measure_map;
    """
    op.execute(sa.text(query))
    # set geometries
    query = """
    UPDATE control_measure_location
    SET geom = (
        SELECT v2_connection_nodes.the_geom
        FROM v2_connection_nodes
        WHERE v2_connection_nodes.id = control_measure_location.connection_node_id
    )
    """
    op.execute(sa.text(query))


def reformat_action_table():
    # Reformat table with action settings to proper csv table:
    # * replace ';' column separators with ','
    # * replace '#' line seperators with '\n'
    # * replace whitespace tabs ('\t') with single whitespace (' ')
    query = """
    UPDATE table_control
    SET action_table = REPLACE(
                          REPLACE(
                              REPLACE(action_table, ';', ','),
                          '#', '\n'),
                      '\t', ' ');
    """
    op.execute(sa.text(query))


def reformat_action_value():
    # Split action_value string (val_1, val_2) into two fields and remove action_value
    # Before: memory_control.action_value = action_value_1, action_value_2
    # After: memory_control.action_value_1 = action_value_1; memory_control.action_value_2 = action_value_2
    query = """
    UPDATE memory_control
    SET action_value_1 = CAST(SUBSTR(action_value, 1, INSTR(REPLACE(REPLACE(REPLACE(action_value, ';', ','), '\t', ' '), '  ', ' '), ',') - 1) AS REAL),
    action_value_2 = CAST(SUBSTR(action_value, INSTR(REPLACE(REPLACE(REPLACE(action_value, ';', ','), '\t', ' '), '  ', ' '), ',') + 1) AS REAL);
    """
    op.execute(sa.text(query))
    remove_column_from_table('memory_control', 'action_value')


def rename_measure_operator(table_name: str):
    op.execute(sa.text(
        f"""
        UPDATE {table_name}
        SET measure_variable = 'water_level'
        WHERE measure_variable = 'waterlevel'
        """
    ))


def remove_column_from_table(table_name: str, column: str):
    with op.batch_alter_table(table_name) as batch_op:
        batch_op.drop_column(column)


def remove_tables(tables: List[str]):
    for table in tables:
        drop_geo_table(op, table)


def make_geom_col_notnull(table_name):
    # Make control_measure_map.geom not nullable by creating a new table with
    # not nullable geometry column, copying the data from control_measure_map
    # to the new table, dropping the original and renaming the new one to control_measure_map
    # For some reason, changing this via batch_op.alter_column does not seem to work

    # Retrieve column names and types from table
    # Note that it is expected that the geometry column is the last column!
    connection = op.get_bind()
    columns = connection.execute(sa.text(f"PRAGMA table_info('{table_name}')")).fetchall()
    col_names = [col[1] for col in columns]
    col_types = [col[2] for col in columns]
    cols = (['id INTEGER PRIMARY KEY'] +
            [f'{cname} {ctype}' for cname, ctype in zip(col_names[:-1], col_types[:-1]) if cname != 'id'] +
            [f'geom {columns[-1][2]} NOT NULL'])
    # Create new table, insert data, drop original and rename to table_name
    temp_name = f'_temp_224_{uuid.uuid4().hex}'
    op.execute(sa.text(f"CREATE TABLE {temp_name} ({','.join(cols)});"))
    op.execute(sa.text(f"INSERT INTO {temp_name} ({','.join(col_names)}) SELECT {','.join(col_names)} FROM {table_name}"))
    drop_geo_table(op, table_name)
    op.execute(sa.text(f"ALTER TABLE {temp_name} RENAME TO {table_name};"))


def fix_geometry_columns():
    GEO_COL_INFO = [(row[0], row[1].name, row[1].type.geometry_type) for row in ADD_COLUMNS
                    if row[1].name == 'geom']
    for table, column, geotype in GEO_COL_INFO:
        make_geom_col_notnull(table)
        migration_query = f"SELECT RecoverGeometryColumn('{table}', '{column}', {4326}, '{geotype}', 'XY')"
        op.execute(sa.text(migration_query))


def update_use_structure_control():
    op.execute("""
        UPDATE simulation_template_settings SET use_structure_control = CASE
            WHEN
                (SELECT COUNT(*) FROM table_control) = 0 AND
                (SELECT COUNT(*) FROM memory_control) = 0 THEN 0
            ELSE use_structure_control
            END;    
    """)


def upgrade():
    # Remove existing tables (outside of the specs) that conflict with new table names
    drop_conflicting(op, list(ADD_TABLES.keys()) + [new_name for _, new_name in RENAME_TABLES])
    # create new tables and rename existing tables
    create_new_tables(ADD_TABLES)
    rename_tables(RENAME_TABLES)
    # add new columns to existing tables
    add_columns_to_tables(ADD_COLUMNS)
    # populate new tables
    remove_invalid_rows_from_v2_control_measure_map()
    populate_control_measure_map()
    populate_control_measure_location()
    remove_orphan_control('table_control')
    remove_orphan_control('memory_control')
    # modify table contents
    reformat_action_table()
    reformat_action_value()
    add_control_geometries('table_control')
    add_control_geometries('memory_control')
    set_geom_for_control_measure_map()
    rename_measure_operator('table_control')
    rename_measure_operator('memory_control')
    move_setting('model_settings', 'use_structure_control',
                 'simulation_template_settings', 'use_structure_control')
    update_use_structure_control()
    remove_tables(DEL_TABLES)
    # Fix geometry columns and also make all but geom column nullable
    fix_geometry_columns()


def downgrade():
    # Not implemented on purpose
    raise NotImplementedError("Downgrade back from 0.3xx is not supported")



import logging
from pathlib import Path
from unittest.mock import call, MagicMock

import pytest

from threedi_schema.application import upgrade_utils
from threedi_schema.application.schema import get_alembic_config
from threedi_schema.application.threedi_database import ThreediDatabase

data_dir = Path(__file__).parent / "data"


def test_progress_handler():
    progress_func = MagicMock()
    expected_calls = [call(100 * i / 5, "Running upgrade") for i in range(5)]
    mock_record = MagicMock(levelno=logging.INFO, percent=10)
    mock_record.getMessage.return_value = "Running upgrade"
    handler = upgrade_utils.ProgressHandler(progress_func, total_steps=5)
    for _ in range(5):
        handler.handle(mock_record)
    assert progress_func.call_args_list == expected_calls


def test_progress_handler_zero_steps():
    progress_func = MagicMock()
    mock_record = MagicMock(levelno=logging.INFO, percent=40)
    mock_record.getMessage.return_value = "Running upgrade"
    expected_calls = 5 * [call(100, "Running upgrade")]
    handler = upgrade_utils.ProgressHandler(progress_func, total_steps=0)
    for _ in range(5):
        handler.handle(mock_record)
    assert progress_func.call_args_list == expected_calls


@pytest.mark.parametrize(
    "target_revision, nsteps_expected", [("0226", 5), ("0200", 0), (None, 0)]
)
def test_get_upgrade_steps_count(target_revision, nsteps_expected):
    schema = ThreediDatabase(data_dir.joinpath("v2_bergermeer_221.sqlite")).schema
    nsteps = upgrade_utils.get_upgrade_steps_count(
        config=get_alembic_config(),
        current_revision=schema.get_version(),
        target_revision=target_revision,
    )
    assert nsteps == nsteps_expected


def test_get_upgrade_steps_count_pre_200(oldest_sqlite):
    schema = oldest_sqlite.schema
    nsteps = upgrade_utils.get_upgrade_steps_count(
        config=get_alembic_config(),
        current_revision=schema.get_version(),
        target_revision="0226",
    )
    assert nsteps == 27


def test_upgrade_with_progress_func(oldest_sqlite):
    schema = oldest_sqlite.schema
    progress_func = MagicMock()
    schema.upgrade(
        revision="0201",
        backup=False,
        upgrade_spatialite_version=False,
        progress_func=progress_func,
    )
    assert progress_func.call_args_list == [
        call(
            0.0,
            "Running upgrade  -> 0200, Create all tables if they do not exist already.",
        ),
        call(
            50.0,
            "Running upgrade 0200 -> 0201, Migration the old friction_type 4 to 2 (MANNING)",
        ),
    ]


def test_upgrade_with_progress_func_latest(oldest_sqlite):
    schema = oldest_sqlite.schema
    progress_func = MagicMock()
    schema.upgrade(
        backup=False,
        upgrade_spatialite_version=False,
        progress_func=progress_func,
        revision="0300",
    )
    progress = [_call.args[0] for _call in progress_func.call_args_list]
    # ensure that the reported upgrade is increasing
    # non-increasing progress indicates multiple initializations
    assert all(x < y for x, y in zip(progress, progress[1:]))



"""Reproject geometries to model CRS

Revision ID: 0230
Revises:
Create Date: 2024-11-12 12:30

"""
import uuid

import sqlalchemy as sa
from alembic import op

from threedi_schema.application.errors import InvalidSRIDException
from threedi_schema.application.schema import get_model_srid
from threedi_schema.migrations.utils import get_crs_info

# revision identifiers, used by Alembic.
revision = "0230"
down_revision = "0229"
branch_labels = None
depends_on = None

GEOM_TABLES = ['boundary_condition_1d', 'boundary_condition_2d', 'channel', 'connection_node', 'measure_location',
               'measure_map', 'memory_control', 'table_control', 'cross_section_location', 'culvert',
               'dem_average_area', 'dry_weather_flow', 'dry_weather_flow_map', 'exchange_line', 'grid_refinement_line',
               'grid_refinement_area', 'lateral_1d', 'lateral_2d', 'obstacle', 'orifice', 'pipe', 'potential_breach',
               'pump', 'pump_map', 'surface', 'surface_map', 'weir', 'windshielding_1d']


def get_geom_type(table_name, geo_col_name):
    # map geometry numbers to names
    geom_type_map = {
        1: 'POINT',
        2: 'LINESTRING',
        3: 'POLYGON',
        4: 'MULTIPOINT',
        5: 'MULTILINESTRING',
        6: 'MULTIPOLYGON',
        7: 'GEOMETRYCOLLECTION',
    }
    connection = op.get_bind()
    # use metadata to determine spatialite version because the geometry type column differs
    srs_wkt_exists = connection.execute(sa.text("select count(name) from pragma_table_info('spatial_ref_sys') where name is 'srs_wkt'")).scalar() == 1
    # spatialite 3
    if srs_wkt_exists:
        return connection.execute(
            sa.text(f"SELECT type from geometry_columns where f_table_name='{table_name}'")).fetchone()[0]
    else:
        geom_type_num = connection.execute(
            sa.text(f"SELECT geometry_type from geometry_columns where f_table_name='{table_name}'")).fetchone()[0]
        return geom_type_map.get(geom_type_num, 'GEOMETRY')

def add_geometry_column(table: str, name: str, srid: int, geometry_type: str):
    # Adding geometry columns via alembic doesn't work
    query = (
        f"SELECT AddGeometryColumn('{table}', '{name}', {srid}, '{geometry_type}', 'XY', 1);")
    op.execute(sa.text(query))


def transform_column(table_name, srid):
    connection = op.get_bind()
    columns = connection.execute(sa.text(f"PRAGMA table_info('{table_name}')")).fetchall()
    # get all column names and types
    skip_cols = ['id', 'geom']
    col_names = [col[1] for col in columns if col[1] not in skip_cols]
    col_types = [col[2] for col in columns if col[1] not in skip_cols]
    # Create temporary table
    temp_table_name = f'_temp_230_{table_name}_{uuid.uuid4().hex}'
    # Create new table, insert data, drop original and rename temp to table_name
    col_str = ','.join(['id INTEGER PRIMARY KEY NOT NULL'] + [f'{cname} {ctype}' for cname, ctype in
                                                              zip(col_names, col_types)])
    op.execute(sa.text(f"CREATE TABLE {temp_table_name} ({col_str});"))
    # Add geometry column with new srid!
    geom_type = get_geom_type(table_name, 'geom')
    add_geometry_column(temp_table_name, 'geom', srid, geom_type)
    # Copy transformed geometry and other columns to temp table
    col_str = ','.join(['id'] + col_names)
    op.execute(sa.text(f"""
        INSERT INTO {temp_table_name} ({col_str}, geom)
        SELECT {col_str}, ST_Transform(geom, {srid}) AS geom FROM {table_name}
        """))
    # Short linestrings may become too short, and invalid, after the transformation
    # By moving the second point 1 centimeter this issue is solved
    if geom_type == "LINESTRING":
        op.execute(sa.text(f"""
            UPDATE {temp_table_name}
            SET geom = MakeLine(
                ST_PointN(geom, 1),
                ST_Translate(ST_PointN(geom, 2), 0.01, 0, 0)
            )
            WHERE ST_Length(geom) = 0
            """))
    # Discard geometry column in old table
    op.execute(sa.text(f"SELECT DiscardGeometryColumn('{table_name}', 'geom')"))
    op.execute(sa.text(f"SELECT DiscardGeometryColumn('{temp_table_name}', 'geom')"))
    # Remove old table
    op.execute(sa.text(f"DROP TABLE '{table_name}'"))
    # Rename temp table
    op.execute(sa.text(f"ALTER TABLE '{temp_table_name}' RENAME TO '{table_name}';"))
    # Recover geometry stuff
    # This gives a bunch of warnings but seems to be needed to fix spatialite stuff
    op.execute(sa.text(f"SELECT RecoverGeometryColumn('{table_name}', "
                       f"'geom', {srid}, '{geom_type}', 'XY')"))
    op.execute(sa.text(f"SELECT CreateSpatialIndex('{table_name}', 'geom')"))
    op.execute(sa.text(f"SELECT RecoverSpatialIndex('{table_name}', 'geom')"))


def prep_spatialite(srid: int):
    conn = op.get_bind()
    has_srid = conn.execute(sa.text(f'SELECT COUNT(*) FROM spatial_ref_sys WHERE srid = {srid};')).fetchone()[0] > 0
    if not has_srid:
        conn.execute(sa.text(f"InsertEpsgSrid({srid})"))


def upgrade():
    # retrieve srid from model settings
    # raise exception if there is no srid, or if the srid is not valid
    srid = get_model_srid(connection=op.get_bind())
    unit, is_projected = get_crs_info(srid)
    if unit != "metre":
        raise InvalidSRIDException(srid, f"the CRS must be in metres, not {unit}")
    if not is_projected:
        raise InvalidSRIDException(srid, "the CRS must be in projected")
    if srid is not None:
        # prepare spatialite databases
        prep_spatialite(srid)
        # transform all geometries
        for table_name in GEOM_TABLES:
            transform_column(table_name, srid)
    else:
        print('Model without geometries and epsg code, we need to think about this')
    # remove crs from model_settings
    with op.batch_alter_table('model_settings') as batch_op:
        batch_op.drop_column('epsg_code')


def downgrade():
    # Not implemented on purpose
    raise NotImplementedError("Downgrade back from 0.3xx is not supported")



import os

from alembic import context
from sqlalchemy import text

import threedi_schema.domain.models  # NOQA needed for autogenerate
from threedi_schema import ThreediDatabase
from threedi_schema.domain import constants
from threedi_schema.domain.models import Base

target_metadata = Base.metadata
config = context.config


def get_url():
    db_url = os.environ.get("DB_URL")
    if not db_url:
        raise RuntimeError(
            "Database URL must be specified using the environment variable DB_URL"
        )
    return db_url


def run_migrations_online():
    """Run migrations in 'online' mode.

    Note: SQLite does not (completely) support transactions, so, backup the
    SQLite before running migrations.
    """
    unsafe = config.attributes.get("unsafe")
    
    engine = config.attributes.get("engine")
    if engine is None:
        engine = ThreediDatabase(get_url()).engine
    

    with engine.connect() as connection:
        # the following 5 lines have been commented out because for some reason the
        # spatialite schema_version doesn't get updated when journal_mode is set to MEMORY
        # TODO: make this work again.
        # if unsafe:
        #     # Speed up by journalling in memory; in case of a crash the database
        #     # will likely go corrupt.
        #     # NB: This setting is scoped to this connection.
        #     connection.execute(text("PRAGMA journal_mode = MEMORY"))

        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            version_table=constants.VERSION_TABLE_NAME,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    raise ValueError("Offline mode is not supported")
else:
    run_migrations_online()



"""Migrate 1d and 2d lateral and boundary condition tables to new schema

Revision ID: 0225
Revises:
Create Date: 2024-08-05 11:22

"""
import uuid
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Tuple

import sqlalchemy as sa
from alembic import op
from geoalchemy2 import load_spatialite
from sqlalchemy import Boolean, Column, Integer, Text
from sqlalchemy.event import listen
from sqlalchemy.orm import declarative_base

from threedi_schema.domain.custom_types import Geometry
from threedi_schema.migrations.utils import drop_conflicting, drop_geo_table

# revision identifiers, used by Alembic.
revision = "0225"
down_revision = "0224"
branch_labels = None
depends_on = None

Base = declarative_base()

data_dir = Path(__file__).parent / "data"


# (source table, destination table)
RENAME_TABLES = [
    ("v2_1d_lateral", "lateral_1d"),
    ("v2_2d_lateral", "lateral_2d"),
    ("v2_1d_boundary_conditions", "boundary_condition_1d"),
    ("v2_2d_boundary_conditions", "boundary_condition_2d"),
]


ADD_COLUMNS = [
    ("lateral_1d", Column("code", Text)),
    ("lateral_1d", Column("display_name", Text)),
    ("lateral_1d", Column("tags", Text)),
    ("lateral_1d", Column("time_units", Text)),
    ("lateral_1d", Column("interpolate", Boolean)),
    ("lateral_1d", Column("offset", Integer)),
    ("lateral_1d", Column("units", Text)),

    ("lateral_2d", Column("code", Text)),
    ("lateral_2d", Column("display_name", Text)),
    ("lateral_2d", Column("tags", Text)),
    ("lateral_2d", Column("time_units", Text)),
    ("lateral_2d", Column("interpolate", Boolean)),
    ("lateral_2d", Column("offset", Integer)),
    ("lateral_2d", Column("units", Text)),

    ("boundary_condition_1d", Column("code", Text)),
    ("boundary_condition_1d", Column("display_name", Text)),
    ("boundary_condition_1d", Column("tags", Text)),
    ("boundary_condition_1d", Column("time_units", Text)),
    ("boundary_condition_1d", Column("interpolate", Boolean)),

    ("boundary_condition_2d", Column("code", Text)),
    # display_name was already added in migration 200
    ("boundary_condition_2d", Column("tags", Text)),
    ("boundary_condition_2d", Column("time_units", Text)),
    ("boundary_condition_2d", Column("interpolate", Boolean)),
]

# Geom columns need to be added using geoalchemy, so therefore that's a separate task
NEW_GEOM_COLUMNS = {
    ("lateral_1d", Column("geom", Geometry("POINT"), nullable=False)),
    ("boundary_condition_1d", Column("geom", Geometry("POINT"), nullable=False)),
}


# old name, new name
# the columns will be renamed using raw sql
# this is because alembic has conniptions whenever you try to batch rename a geometry column
RENAME_COLUMNS = {
    "boundary_condition_1d": [
        ("boundary_type", "type"),
    ],
    "boundary_condition_2d": [
        ("boundary_type", "type"),
        ("the_geom", "geom"),
    ],
    "lateral_2d": [
        ("the_geom", "geom"),
    ],
}


DEFAULT_VALUES = {
    "lateral_1d": {
        "time_units": "'minutes'",
        "interpolate": "0",  # false
        "offset": "NULL",
        "units": "'m3/s'"
    },
    "lateral_2d": {
        "time_units": "'minutes'",
        "interpolate": "0",  # false
        "offset": "NULL",
        "units": "'m3/s'"
    },
    "boundary_condition_1d": {
        "time_units": "'minutes'",
        "interpolate": "1",  # true
    },
    "boundary_condition_2d": {
        "time_units": "'minutes'",
        "interpolate": "1",  # true
    },
}

GEOMETRY_TYPES = {"lateral_2d": "POINT",
                  "lateral_1d": "POINT",
                  "boundary_condition_1d": "POINT",
                  "boundary_condition_2d": "LINESTRING"}


def rename_tables(table_sets: List[Tuple[str, str]]):
    # no checks for existence are done, this will fail if a source table doesn't exist
    connection = op.get_bind()
    spatialite_version = connection.execute(sa.text("SELECT spatialite_version();")).fetchall()[0][0]
    for src_name, dst_name in table_sets:
        if spatialite_version.startswith('5'):
            op.execute(sa.text(f"SELECT RenameTable(NULL, '{src_name}', '{dst_name}');"))
        else:
            op.rename_table(src_name, dst_name)


def create_new_tables(new_tables: Dict[str, sa.Column]):
    # no checks for existence are done, this will fail if any table already exists
    for table_name, columns in new_tables.items():
        op.create_table(table_name, sa.Column("id", sa.Integer(), primary_key=True),
                        *columns)


def add_columns_to_tables(table_columns: List[Tuple[str, Column]]):
    # no checks for existence are done, this will fail if any column already exists
    for dst_table, col in table_columns:
        if isinstance(col.type, Geometry):
            add_geometry_column(dst_table, col)
        else:
            with op.batch_alter_table(dst_table) as batch_op:
                batch_op.add_column(col)


def add_geometry_column(table: str, geocol: Column):
    # Adding geometry columns via alembic doesn't work
    # https://postgis.net/docs/AddGeometryColumn.html
    geotype = geocol.type
    query = (
        f"SELECT AddGeometryColumn('{table}', '{geocol.name}', {geotype.srid}, '{geotype.geometry_type}', 'XY', {int(not geocol.nullable)});")
    op.execute(sa.text(query))


def rename_columns(table_name: str, columns: List[Tuple[str, str]]):
    # no checks for existence are done, this will fail if table or any source column doesn't exist
    connection = op.get_bind()
    old_columns_result = connection.execute(sa.text(f"PRAGMA table_info('{table_name}')")).fetchall()
    old_columns = []
    for value_list in old_columns_result:
        old_columns.append({"name": value_list[1], "type": value_list[2]})

    columns_dict = dict(columns)

    old_columns_list = [entry["name"] for entry in old_columns]
    if not all(e in old_columns_list for e in columns_dict):
        raise ValueError(f"Cannot rename columns {columns_dict.keys()} in table {table_name}; table does not contain all these columns")
    new_columns = deepcopy(old_columns)
    for i in range(len(new_columns)):
        if new_columns[i]["name"] in columns_dict.keys():
            new_columns[i]["name"] = columns_dict[new_columns[i]["name"]]

    new_columns_list = [entry["name"] for entry in new_columns]

    new_columns_list_sql_formatted = []
    for entry in new_columns:
        entry_string = f"{entry['name']} {entry['type']}"
        if entry['name'] == "id":
            entry_string += f" PRIMARY KEY"
        # in the new database schema only geometries and id will have NOT NULL constraints
        if entry['name'] in ["geom", "id"]:
            entry_string += f" NOT NULL"
        new_columns_list_sql_formatted.append(entry_string)

    temp_name = f'_temp_225_{uuid.uuid4().hex}'
    create_table_query = f"""CREATE TABLE {temp_name} ({', '.join(new_columns_list_sql_formatted)});"""
    op.execute(sa.text(create_table_query))
    op.execute(sa.text(f"INSERT INTO {temp_name} ({','.join(new_columns_list)}) SELECT {','.join(old_columns_list)} from {table_name};"))
    drop_geo_table(op, table_name)
    op.execute(sa.text(f"ALTER TABLE {temp_name} RENAME TO {table_name};"))

    if table_name is not GEOMETRY_TYPES:
        op.execute(sa.text(f"""SELECT RecoverGeometryColumn('{table_name}', 'geom', 4326, '{GEOMETRY_TYPES[table_name]}', 'XY')"""))



def copy_v2_geometries_from_connection_nodes_by_id(dest_table: str, dest_column: str):
    query = (
        f"""
        UPDATE {dest_table}
        SET {dest_column} = (
            SELECT the_geom
            FROM v2_connection_nodes
            WHERE {dest_table}.connection_node_id = v2_connection_nodes.id
        );
        """
    )
    op.execute(sa.text(query))


def populate_table(table: str, values: dict):
    """Populate SQL columns with values"""
    # convert {a: b, c: d} to "a=b, c=d" for the query
    sql_formatted_columns = ', '.join('{} = {}'.format(key, value) for key, value in values.items())
    # then insert it into the query
    query = f"""UPDATE {table} SET {sql_formatted_columns};"""
    op.execute(sa.text(query))


def upgrade():
    # Drop tables that conflict with new table names
    drop_conflicting(op, [new_name for _, new_name in RENAME_TABLES])

    # rename existing tables
    rename_tables(RENAME_TABLES)

    # add new columns to existing tables
    add_columns_to_tables(ADD_COLUMNS)

    # rename columns in renamed tables
    for table_name, columns in RENAME_COLUMNS.items():
        rename_columns(table_name, columns)

    # add geometry columns after renaming columns
    # to not needlessly trigger RecoverGeometryColumn
    add_columns_to_tables(NEW_GEOM_COLUMNS)

    # recover geometry column data from connection nodes
    for table, column in (
        ("lateral_1d", "geom"),
        ("boundary_condition_1d", "geom")
    ):
        copy_v2_geometries_from_connection_nodes_by_id(dest_table=table, dest_column=column)

    # populate new columns in tables
    for key, value in DEFAULT_VALUES.items():
        populate_table(table=key, values=value)


def downgrade():
    # Not implemented on purpose
    raise NotImplementedError("Downgrade back from 0.3xx is not supported")
